package tests;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.List;

import org.springframework.context.ApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import service.IService;
import entites.Personne;

public class TestNG {

	// couche service
	private IService service;

	@BeforeClass
	public void init() {
		// log
		log("init");
		// configuration de l'application
		ApplicationContext ctx = new ClassPathXmlApplicationContext("spring-config.xml");
		// couche service
		service = (IService) ctx.getBean("service");
	}

	@BeforeMethod
	public void setUp() throws ParseException {
		// on vide la base
		clean();
		// on la remplit
		fill();
	}

	// logs
	private void log(String message) {
		System.out.println("----------- " + message);
	}

	// affichage contenu table
	private void dump() {
		log("dump");
		System.out.format("[personnes]%n");
		for (Personne p : service.getAll()) {
			System.out.println(p);
		}
	}

	// remplissage table
	public void fill() throws ParseException {
		log("fill");
		// cration personnes
		Personne p1 = new Personne("p1", "Paul", new SimpleDateFormat("dd/MM/yy").parse("31/01/2000"), true, 2);
		Personne p2 = new Personne("p2", "Sylvie", new SimpleDateFormat("dd/MM/yy").parse("05/07/2001"), false, 0);
		// qu'on sauvegarde
		service.saveArray(new Personne[] { p1, p2 });
	}

	// supression lments de la table
	public void clean() {
		log("clean");
		for (Personne p : service.getAll()) {
			service.deleteOne(p.getId());
		}
	}

	@Test()
	public void test01() {
		log("test1");
		dump();
		// liste des personnes
		List<Personne> personnes = service.getAll();
		assert 2 == personnes.size();
	}

	@Test()
	public void test02() {
		log("test2");
		// recherche de personnes par leur nom
		List<Personne> personnes = service.getAllLike("p1%");
		assert 1 == personnes.size();
		Personne p1 = personnes.get(0);
		assert "Paul".equals(p1.getPrenom());
	}

	@Test()
	public void test03() throws ParseException {
		log("test3");
		// cration d'une nouvelle personne
		Personne p3 = new Personne("p3", "x", new SimpleDateFormat("dd/MM/yy").parse("05/07/2001"), false, 0);
		// on la persiste
		service.saveOne(p3);
		// on la redemande
		Personne loadedp3 = service.getOne(p3.getId());
		// on l'affiche
		System.out.println(loadedp3);
		// vrification
		assert "p3".equals(loadedp3.getNom());
	}

	@Test()
	public void test04() throws ParseException {
		log("test4");
		// on charge la personne p1
		List<Personne> personnes = service.getAllLike("p1%");
		Personne p1 = personnes.get(0);
		// on l'affiche
		System.out.println(p1);
		// on vrifie
		assert "p1".equals(p1.getNom());
		int version1 = p1.getVersion();
		// on modifie le prnom
		p1.setPrenom("x");
		// on sauvegarde
		service.updateOne(p1);
		// on recharge
		p1 = service.getOne(p1.getId());
		// on l'affiche
		System.out.println(p1);
		// on vrifie que la version a t incrmente
		assert (version1 + 1) == p1.getVersion();

	}

	@Test()
	public void test05() {
		log("test5");
		// on charge la personne p2
		List<Personne> personnes = service.getAllLike("p2%");
		Personne p2 = personnes.get(0);
		// on l'affiche
		System.out.println(p2);
		// on vrifie
		assert "p2".equals(p2.getNom());
		// on supprime la personne p2
		service.deleteOne(p2.getId());
		// on la recharge
		p2 = service.getOne(p2.getId());
		// on vrifie qu'on a obtenu un pointeur null
		assert null == p2;
		// on affiche la table
		dump();
	}

	@Test()
	public void test06() throws ParseException {
		log("test6");
		// on cre un tableau de 2 personnes de mme nom (enfreint la rgle d'unicit du nom)
		Personne[] personnes = { new Personne("p3", "x", new SimpleDateFormat("dd/MM/yy").parse("31/01/2000"), true, 2),
				new Personne("p4", "x", new SimpleDateFormat("dd/MM/yy").parse("31/01/2000"), true, 2),
				new Personne("p4", "x", new SimpleDateFormat("dd/MM/yy").parse("31/01/2000"), true, 2)};
		// on sauvegarde ce tableau - on doit obtenir une exception et un rollback
		boolean erreur = false;
		try {
			service.saveArray(personnes);
		} catch (RuntimeException e) {
			erreur = true;
		}
		// dump
		dump();
		// vrifications
		assert erreur;
		// recherche personne de nom p3
		List<Personne> personnesp3 = service.getAllLike("p3%");
		assert 0 == personnesp3.size();
		// dump
		dump();
	}

	@Test()
	public void test07() {
		log("test7");
		// test optimistic locking
		// on charge la personne p1
		List<Personne> personnes = service.getAllLike("p1%");
		Personne p1 = personnes.get(0);
		// on l'affiche
		System.out.println(p1);
		// on augmente son nbre d'enfants
		int nbEnfants1 = p1.getNbenfants();
		p1.setNbenfants(nbEnfants1 + 1);
		// on sauvegarde p1
		Personne newp1 = service.updateOne(p1);
		assert (nbEnfants1 + 1) == newp1.getNbenfants();
		System.out.println(newp1);
		// on sauvegarde une deuxime fois - on doit avoir une exception car p1 n'a plus la bonne version
		// c'est newp1 qui l'a
		boolean erreur = false;
		try {
			service.updateOne(p1);
		} catch (RuntimeException e) {
			erreur = true;
		}
		// vrification
		assert erreur;
		// on augmente le nbre d'enfants de newp1
		int nbEnfants2 = newp1.getNbenfants();
		newp1.setNbenfants(nbEnfants2 + 1);
		// on sauvegarde newp1
		service.updateOne(newp1);
		// on recharge
		p1 = service.getOne(p1.getId());
		// on vrifie
		assert (nbEnfants1 + 2) == p1.getNbenfants();
		System.out.println(p1);
	}

	@Test()
	public void test08() {
		log("test8");
		// test rollback sur updateArray
		// on charge la personne p1
		List<Personne> personnes = service.getAllLike("p1%");
		Personne p1 = personnes.get(0);
		// on l'affiche
		System.out.println(p1);
		// on augmente son nbre d'enfants
		int nbEnfants1 = p1.getNbenfants();
		p1.setNbenfants(nbEnfants1 + 1);
		// on sauvegarde 2 modifications dont la 2ime doit chouer (personne mal initialise)
		//  cause de la transaction, les deux doivent alors tre annules
		boolean erreur = false;
		try {
			service.updateArray(new Personne[] { p1, new Personne() });
		} catch (RuntimeException e) {
			erreur = true;
		}
		// vrifications
		assert erreur;
		// on recharge la personne p1
		personnes = service.getAllLike("p1%");
		p1 = personnes.get(0);
		// son nbre d'enfants n'a pas du changer
		assert nbEnfants1 == p1.getNbenfants();
	}

	@Test()
	public void test09() {
		log("test9");
		// test rollback sur deleteArray
		// dump
		dump();
		// on charge la personne p1
		List<Personne> personnes = service.getAllLike("p1%");
		Personne p1 = personnes.get(0);
		// on l'affiche
		System.out.println(p1);
		// on fait 2 suppressions dont la 2ime doit chouer (personne inconnue)
		//  cause de la transaction, les deux doivent alors tre annules
		boolean erreur = false;
		try {
			service.deleteArray(new Personne[] { p1, new Personne() });
		} catch (RuntimeException e) {
			erreur = true;
		}
		// vrifications
		assert erreur;
		// on recharge la personne p1
		personnes = service.getAllLike("p1%");
		// vrification
		assert 1 == personnes.size();
		// dump
		dump();
	}

	// optimistic locking - accs multi-threads
	@Test()
	public void test10() throws Exception {
		// ajout d'une personne
		Personne p3 = new Personne("X", "X", new SimpleDateFormat("dd/MM/yyyy").parse("01/02/2006"), true, 0);
		service.saveOne(p3);
		int id3 = p3.getId();
		// cration de N threads de mise  jour du nombre d'enfants
		final int N = 20;
		Thread[] taches = new Thread[N];
		for (int i = 0; i < taches.length; i++) {
			taches[i] = new ThreadMajEnfants("thread n " + i, service, id3);
			taches[i].start();
		}
		// on attend la fin des threads
		for (int i = 0; i < taches.length; i++) {
			taches[i].join();
		}
		// on rcupre la personne
		p3 = service.getOne(id3);
		// elle doit avoir N enfants
		assert N == p3.getNbenfants();
		// suppression personne p3
		service.deleteOne(p3.getId());
		// vrification
		p3 = service.getOne(p3.getId());
		// on doit avoir un pointeur null
		assert p3 == null;
	}
}
